;; PSIMETRICA:  Phonological SIMilarity METRIC Analysis
;; (c) 1998-2004 Shane T. Mueller
;; stmuelle@indiana.edu
;; More info at: http://mypages.iu.edu/~stmuelle
;; May not be copied or distributed without permission.
;; Tools and routines for measuring phonological similarity between words.
;;
;;  similarity.lsp
;;  This file defines the feature set and  contains functions for  
;;  measuring the similarity between phonemes, syllables and words.



;;;==========================================================================
;;;        *phonemes* contains the sub-symbolic description of phonemes
;;;==========================================================================
;;;This data is derived from chart on pages 176-177 of
;;;_The Sound Pattern of English_, Noam Chomsky, 1968.
;;;The first 2-3 letter code is the phoneme label, and the rest
;;;of the vector represents the speech features that characterize
;;;that phoneme.  They are, in order:  vocalic, consonantal, high, back,low,
;;;anterior, coronal, round, tense, voice, continuant, nasal, strident.
;;;As Chomsky notes, anterior and coronal replace labial, dental, paleto-alveolar, and velar
;;;The Xs represent unspecified values.  For the moment, they remain unspecified,
;;;but could be determined given phonologic rules and specific words.

(defvar *phonemes-nc* '((i1  (+ - + - - - - - + x x x x))
                        (i2  (+ - + + - - - - + x x x x))
                        (u1  (+ - + + - - - + + x x x x))
                        (e1  (+ - - - - - - - + x x x x))
                        (o1  (+ - - + - - - + + x x x x))
                        (ae1 (+ - - - + - - - + x x x x))
                        (a+  (+ - - + + - - - + x x x x))
                        (-ee (+ - - - + - - + + x x x x))
                        (-e  (+ - - + + - - + + x x x x))
                        (i3  (+ - + - - - - - - x x x x))
                        (u2  (+ - + + - - - + - x x x x))
                        (e2  (+ - - - - - - - - x x x x))
                        (u3  (+ - - + - - - - - x x x x))
                        (o2  (+ - - + - - - + - x x x x))
                        (ae2 (+ - - - + - - - - x x x x))
                        (-e2 (+ - - + + - - + - x x x x))
                        (y   (- - + - - - - - - x x x x))
                        (w   (- - + + - - - + - x x x x))
                        (eps (- - - - - - - - - x x x x))
                        (r   (+ + - - - - + x x + + - -))
                        (l   (+ + - - - + + x x + + - -))
                        (p   (- + - - - + - x x - - - -))
                        (b   (- + - - - + - x x + - - -))
                        (f   (- + - - - + - x x - + - +))
                        (v   (- + - - - + - x x + + - +))
                        (m   (- + - - - + - x x + - + -))
                        (t   (- + - - - + + x x - - - -))
                        (d   (- + - - - + + x x + - - -))
                        (th  (- + - - - + + x x - + - -))
                        (thz (- + - - - + + x x + + - -))
                        (n   (- + - - - + + x x + - + -))
                        (s   (- + - - - + + x x - + - +))
                        (z   (- + - - - + + x x + + - +))
                        (c   (- + - - - + + x x - - - +))
                        (cc  (- + + - - - + x x - - - +))
                        (jj  (- + + - - - + x x + - - +))
                        (ss  (- + + - - - + x x - + - +))
                        (zz  (- + + - - - + x x + + - +))
                        (k   (- + + + - - - - x - - - -))
                        (g   (- + + + - - - - x + - - -))
                        (x   (- + + + - - - - x - + - -))
                        (ng  (- + + + - - - x x + - + -))
                        (h   (- - - - + - - x x - + - -))
                        (kw  (- + + + - - - + x - - - -))
                        (gw  (- + + + - - - + x + - - -))
                        (xw  (- + + + - - - + x - + - -))


                        ))

;                    symbol  feature-vector                IPA symbol......MRC.......Example
(defvar *phonemes*    '((ee  (+ - + - - - - - + x x x x))  ;lower-case i...i.........steam
                        (i   (+ - + - - - - - - x x x x))  ;block i........I.........pits
                        (a_  (+ - - - - - - - + x x x x))  ;e..............eI........late
                        (e   (+ - - - - - - - - x x x x))  ;epsilon........e.........mess
                        (ae  (+ - - - + - - - + x x x x))  ;ae.............&.........fat
                        (u   (+ - + + - - - - - x x x x))  ;ibar........... .........just
                        (@   (+ - - + - - - - - x x x x))  ;inverse e...... .........burn, eatEn, Above
                        (uu  (+ - - + + - - - - x x x x))  ;inverse v......V.........rUbber,abOve
                        (ah  (+ - - + + - - - + x x x x))  ;a..............A.........hAlf
                        (u_  (+ - + + - - - + + x x x x))  ;u..............u.........crUise
                        (oo  (+ - + + - - - + - x x x x))  ;block U........U.........cOOk
                        (o_  (+ - - + - - - + + x x x x))  ;o..............O.........cOAt
                        (aw  (+ - - + + - - + - x x x x))  ;backwards c....0 (zero)..bAWl
                        (y   (- - + - - - - - - x x x x))  ;j..............j.........Yell
                        (w   (- - + + - - - + - x x x x))  ;w..............w.........Wool
                        (r   (+ + - - - - + x x + + - -))  ;r..............r.........Rat
                        (l   (+ + - - - + + x x + + - -))  ;l..............l.........Loin
                        (h   (- - - - + - - x x - + - -))  ;h..............h.........Hot
                        (p   (- + - - - + - x x - - - -))  ;p..............p.........Pot
                        (b   (- + - - - + - x x + - - -))  ;b..............b.........But
                        (t   (- + - - - + + x x - - - -))  ;t..............t.........Top
                        (d   (- + - - - + + x x + - - -))  ;d..............d.........Dog
                        (ch  (- + + - - - + x x - - - +))  ;tSigma.........tS........CHurCH
                        (j   (- + + - - - + x x + - - +))  ;d3.............dZ........Jowls
                        (k   (- + + + - - - - x - - - -))  ;k..............k.........Cramp
                        (g   (- + + + - - - - x + - - -))  ;g..............g.........Goon
                        (f   (- + - - - + - x x - + - +))  ;f..............f.........Feeler
                        (v   (- + - - - + - x x + + - +))  ;v..............v.........Vain
                        (th  (- + - - - + + x x - + - -))  ;theta..........T.........THought
                        (thz (- + - - - + + x x + + - -))  ;delta..........D.........THis
                        (s   (- + - - - + + x x - + - +))  ;s..............s.........Sign
                        (z   (- + - - - + + x x + + - +))  ;z..............z.........Zircon
                        (sh  (- + + - - - + x x - + - +))  ;sigma..........S.........SHeet
                        (zh  (- + + - - - + x x + + - +))  ;3..............Z.........viSIon
                        (m   (- + - - - + - x x + - + -))  ;m..............m.........MiMe
                        (n   (- + - - - + + x x + - + -))  ;n..............n.........Numb
                        (ng  (- + + + - - - x x + - + -))  ;eta............9.........youNG
                        (X   (1 1 1 1 1 1 1 1 1 1 1 1 1))  ;default placeholder
                        (XX  (0 0 0 0 0 0 0 0 0 0 0 0 0))  ;default nil phoneme
                        ;total:  37 phonemes +2 placeholder
                        ))



(defvar *consonants* '(h p b t d ch j k g f v th thz m n ng))
(defvar *lgs* '(y w r l))
(defvar *ss*  '(s z sh zh))

(defvar *distance-scale* 1)
(defvar *vowel-strength* 1)
(defvar *consonant-strength* 1)
(defvar *temperature 1)
;(defvar *average-consonant-distance* (mean-distance *consonants*))
;(defvar *average-lg-distance* (roundoff (/ (+ (pd 'w 'y)   (pd 'r 'l)) 2)))
;(defvar *average-ss-distance*  (mean-distance *ss*))
(defvar *rhyme-weight* 0)
(defvar *k-weight 1)

(setf *distance* .37)
(setf *average-consonant-distance*  .37); *distance*)
(setf *average-lg-distance*        .37) ;  *distance*)
(setf *average-ss-distance*        .37);   *distance*)


;;;*recog-mem* is essentially ltm for actual words. It represents
;;;words on list--those with a great deal of current familiarity.  There
;;;may be a place for a larger lexicon of all possible words that
;;;more acts automatically,
;;;with a comparison to *recog-mem* being an active comparison.
;;;This may be required to account for non-word performance.
;;;Question:  recognition memory decay????
;;;LISTS L1 TO C4 WERE PHONEMIFIED USING WEBSTER'S NEW COLLEGIATE DICTIONARY (G.C. MERRIAM COMPANY,
;;;SPRINGFIELD, MA, USA,1981.
;;;DECEMBER 13, 1998.

;;;Syllable transitions are as follows: single r & l are placed 
;;; at end of first syllable with (X) in second syllable start
;;; all other single consonants go in at start of second syllable 
;;; start with an (X) in first syllable ending


;;;============================================================================
;;;            VECTORIFY functions return vector description of phoneme
;;;============================================================================


(defun vectorify-word (word-as-list-of-syllables &optional word-so-far)
  (cond ((endp word-as-list-of-syllables) word-so-far)
        (t      (vectorify-word (rest word-as-list-of-syllables)
                                (append word-so-far
                                        (list (vectorify-syllable (first word-as-list-of-syllables))))))))




(defun vectorify-syllable (syllable)
  (list
   (first syllable)
   (vectorify-segment (second syllable))
   (vectorify-segment (third syllable))
   (vectorify-segment (fourth syllable))))

(defun vectorify-segment (syllable-as-list-of-phonemes &optional syllable-so-far)
  (cond ((endp syllable-as-list-of-phonemes) syllable-so-far)
        (t     (vectorify-segment (rest syllable-as-list-of-phonemes)
                                  (append  syllable-so-far
                                           (list (vectorify-phoneme (first syllable-as-list-of-phonemes))))))))
(defun vectorify-phoneme (phoneme)
  (second (assoc phoneme *phonemes*)))

;;;============================================================================
;;;   LIST operations:  Converts lists to vectors and back again
;;;============================================================================


;;;This changes a list of words into a list of vectors
(defun vectorify-list (list &optional (new-list nil))
  (cond ((endp list) new-list)
        (t (vectorify-list (rest list)
                           (append new-list
                                   (list (vectorify-word (first list))))))))

;;============================================================================
;;;    Converts possibly unfaithful phoneme description to word on list
;;;           These functions "redintegrate" a degraded trace
;;;============================================================================





;;;A value-list is constructed where a word is assigned points: 1 point
;;;for every phoneme that doesn't
;;;match, 1 point for each difference in length
;;;This function takes a vector-based word and a phoneme-based list (and so
;;; can take degraded traces)


(defun make-value-list (word-trace recognition-memory)
  (make-value-list-2 word-trace
                     (append recognition-memory
                             (list (subst 'xx '(0 0 0 0 0 0 0 0 0 0 0 0 0)
                                          (destroy-word word-trace 1) :test #'equal)))))

(defun make-value-list-2 (word-trace recognition-memory &optional (value-list nil) )
  (cond ((endp recognition-memory) value-list)
        (t (make-value-list-2 word-trace
                              (rest recognition-memory)
                              (append value-list
                                      (list (list (first recognition-memory)
                                                  (evaluate-dissimilarity word-trace
                                                                          (first recognition-memory)))))))))


;;;phoneme-word is a feature vector.  target-word is a phoneme vector
(defun evaluate-dissimilarity (phoneme-word target-word)
  (evaluate-similarity phoneme-word target-word))


(defun find-total-probability (value-list &optional (total 0))
  (cond ((endp value-list) total)
        (t (find-total-probability (rest value-list)
                                   (+ total
                                      (e-to-minus-x (second (first value-list))))))))



;;;the following requires the use of a choice rule to translate distance into probability
;;;it takes a value-list: a list of similarities, and converts them to probability using
;;;Luce's choice model

(defun make-probability-list (value-list)
  (make-probability-list-alt value-list (find-total-probability value-list) nil))

(defun make-probability-list-alt (value-list total-prob probability-list)
  (cond ((endp value-list) probability-list)
        (t (make-probability-list-alt (rest value-list)
                                      total-prob
                                      (append probability-list
                                              (list (list (first (first value-list))
                                                          (/ (e-to-minus-x (second (first value-list)))
                                                             total-prob))))))))




;;;============================================================================
;;;                         -------SIMILARITY----------
;;;  Construct words on list directly from memory vector ala Nairne
;;;  Actually a _DIS_similarity measure: 0 is the most similar
;;;============================================================================
;;;the result of this creates a value-list that can be fed through the make-probability-list
;;;and choose-word functions

(defun make-similarity-list (word-feature-vector ltm-list &optional (similarity nil))
  (cond ((endp ltm-list) similarity)
        (t (make-similarity-list word-feature-vector
                                 (rest ltm-list)
                                 (append similarity
                                         (list (list (first ltm-list)
                                                     (evaluate-similarity word-feature-vector
                                                                          (first ltm-list)))))))))

;;;============================================================================
;; This function measures similarity
;; as of 9/2000, both inputs are feature-based, (previously, only word-feature-vector was, to mimic degradation)
;; it is cleaner if both enter as feature-based vectors.
;;
;;;============================================================================

(defun evaluate-similarity (word-feature-vector target-word-vector)
  (let ((num-syllables  (max (length word-feature-vector)
                             (length target-word-vector))))
    (evaluate-similarity-alt (make-length-equal word-feature-vector target-word-vector)
                             (make-length-equal target-word-vector word-feature-vector)
                             num-syllables)))

;*********************************************************************************************
;   New versian of evaluate-similarity-alt that uses mapcar logic instead of recursion
;   It is marginally faster & smaller than one now found in obsolete file
;*********************************************************************************************
(defun evaluate-similarity-alt (word-feature-vector target-word-vector num-syllables)
  (/ (apply #'+                                    ;add up syllable distances and divide by length
            (mapcar #'(lambda (x y)                ;generate a list of syllable distances
                        (syllable-distance x y))
              word-feature-vector
              target-word-vector)     
      )
      num-syllables))
     
     
;;;============================================================================
;;;   This keeps the different parts of the syllable distinct
;;;============================================================================     
 (defun evaluate-similarity-list (word-feature-vector target-feature-vector)
;   (print "EVALUATE-SIMILARITY-LIST")
   (let ((num-syllables  (max (length word-feature-vector)
                             (length target-feature-vector))))
 ;  (print (list "Max" num-syllables word-feature-vector target-feature-vector))
    (collapse-across-syllables 
    (mapcar #'(lambda (l) (apply #'syllable-distance-list l))
      (transpose (find-minimum-similarity 
                  (assess-potential-similarities (make-n-longer
                                                  word-feature-vector 
                                                  num-syllables)
                                                 (make-n-longer 
                                                  target-feature-vector 
                                                  num-syllables)
                                                 ))))
    )))

;**************************************************************************
;      This function takes two list of potential words (generated by make-n-longer)
;      One list is necessarily one item long (if this changes, this function must be rewritten)
;      It returns a list of all potential pairwise comparisons, with the mean distance attached:
;        e.g.: ((word1 word2a distance) (word1 word2b distance) .....)
; one of the inputs has only one element, because 
; the shorter word is made longer than the long word.
;
;**************************************************************************

(defun assess-potential-similarities (word-feature-vector target-feature-vector)
 ; (pprint (list "----assess-potential-similarities" word-feature-vector target-feature-vector))
  (cond ((and (= (length word-feature-vector) 1)
               (= (length target-feature-vector) 1)) ;If the lengths are equal to 1, just return the pair
         (LIST (list 0 (first word-feature-vector) (first target-feature-vector))))
        
        ((= (length word-feature-vector) 1)  ;;;;;;   word-feature-vector is the longer word    
      
         (mapcar `(lambda (x) 
                    ;;; return a list consisting of (word1 word2 distance) units
                    (list (evaluate-similarity (first (quote ,word-feature-vector)) x) 
                          (first (quote ,word-feature-vector))    ;      
                          x                                       ;   where distance is the average of all parts 
                          ))           
           target-feature-vector)) 
        (t                    ;;;;;;;   target-word-vector is the longer word
      
         (mapcar `(lambda (x) 
                    
                    ;   return a list consisting of (word1 word2 distance) units
                    (list (evaluate-similarity x (first (quote ,target-feature-vector)) ) 
                          x 
                          (first (quote ,target-feature-vector))     ;   where distance is the average of all parts 
                          ))           
          word-feature-vector))
        ))
            

;;; This takes a list of (similarity word1 word2) elements and returns the first (word1 word2) list
;;;  with the smallest similarity

(defun find-minimum-similarity (sim-list)
 ; (PRINT (list "FIND-MINIMUM-SIMILARITY" SIM-LIST))
  (if (= (length sim-list) 1)
      (rest (first sim-list))
    (if (> (first (first sim-list))
           (first (second sim-list)))
        (find-minimum-similarity (rest sim-list))
      (find-minimum-similarity (append (list (first sim-list)) (rest (rest sim-list)))))))




;;;------------------------------------------------------------------------------------------------
;;;                   Syllable-distance
;;;------------------------------------------------------------------------------------------------

(defun syllable-distance (syllable-vector-1 syllable-vector-2)
  (let
    (
     ;(stress (*  (/ (+ (first syllable-vector-1)
     ;             (first syllable-vector-2))
     ;         2)))
     ;(stress 1)
    )
    (/  (+ (segment-o-distance (second syllable-vector-1) (second syllable-vector-2))
          (* (+ (* (segment-v-distance (third syllable-vector-1) (third syllable-vector-2))
                   *vowel-strength*)
                (segment-c-distance (last syllable-vector-1) (last syllable-vector-2)))
             *rhyme-weight*))


       (+ 1
          (* *rhyme-weight*
             (+ *vowel-strength* 1))))))         ;1 for each consonant cluster, and vowel-strength for vowel



;;This just produces the unweighted similarities of phoneme clusters and stress elements
(defun syllable-distance-list (syllable-vector-1 syllable-vector-2)
  (list

   (segment-o-distance (second syllable-vector-1) (second syllable-vector-2))
   (segment-v-distance (third syllable-vector-1) (third syllable-vector-2))
   (segment-c-distance (fourth syllable-vector-1) (fourth syllable-vector-2))
   (stress-distance (first syllable-vector-1) (first syllable-vector-2))
   (/ (+ (cross-consonant-distance (second syllable-vector-1)
                                   (fourth syllable-vector-2))
         (cross-consonant-distance (second syllable-vector-2)
                                   (fourth syllable-vector-2)))
      2.0)
       
   )
  )                                                      
      
      
      
      
 ;;counts the relevant number of phonemes in any pair of phonemes
(defun rel-no-of-phonemes (s-1 s-2 &optional (n 0))
  (if (endp s-1)
    n
    (rel-no-of-phonemes
     (rest s-1)
     (rest s-2)
     (+ n
        (if (and (equal (first s-1) '(1 1 1 1 1 1 1 1 1 1 1 1 1))
                 (equal (first s-2) '(1 1 1 1 1 1 1 1 1 1 1 1 1)))
          0
          1)))))


;;calculates segment distance for onset--already  parsed into 3 parts

(defun segment-o-distance (syllable-vector-1 syllable-vector-2)
  (if (not (and (= 3 (length syllable-vector-1))
                (= 3 (length syllable-vector-2))))
    (progn ;(print "using old distance method")
           (segment-c-distance syllable-vector-1 syllable-vector-2))

    (/ (+  (segment-distance-alt (list (first syllable-vector-1))
                                 (list (first syllable-vector-2))
                                 *average-ss-distance*)
           (segment-distance-alt  (list (second syllable-vector-1))
                                  (list (second syllable-vector-2))
                             *average-consonant-distance*)
           (segment-distance-alt  (list (third syllable-vector-1))
                                  (list (third syllable-vector-2))
                             *average-lg-distance*))
       (max (rel-no-of-phonemes syllable-vector-1 syllable-vector-2)
            1))))




;;calculates segment distance by repeating the shorter segment
;;appropriate for vowels

(defun segment-v-distance (syllable-vector-1 syllable-vector-2)
  (segment-distance-alt (make-length-equal syllable-vector-1 syllable-vector-2)
                        (make-length-equal syllable-vector-2 syllable-vector-1)))


;;calculates segment distance by finding minimum of the possible alignments
;;appropriate for codas

(defun segment-c-distance (segment-vector-1 segment-vector-2)
  (cond ((= (length segment-vector-1) (length segment-vector-2)) (segment-distance-alt
                                                                  segment-vector-1
                                                                  segment-vector-2))
        ((< (length segment-vector-1) (length segment-vector-2))  (segment-c-distance-alt
                                                                   segment-vector-1
                                                                   segment-vector-2))
        (t                                                        (segment-c-distance-alt
                                                                   segment-vector-2
                                                                   segment-vector-1))))




(defun segment-distance-alt (syllable-vector-1 syllable-vector-2 &optional
                                               (null-distance *average-consonant-distance*)
                                               (distance 0)
                                               (len 0))
   (cond ((endp syllable-vector-1) (/ distance len))
        (t                        (segment-distance-alt (rest syllable-vector-1)
                                                        (rest syllable-vector-2)
                                                        null-distance
                                                        (+ distance
                                                           (phoneme-distance (first syllable-vector-1)
                                                                             (first syllable-vector-2)
                                                                             null-distance))
                                                        (+ len 1)))))



;;;Adds null phonemes to either the front or back to make lengths equal
(defun segment-c-distance-alt (s-1 s-2)
  (let* ((dif (- (length s-2) (length s-1)))
         (x '((1 1 1 1 1 1 1 1 1 1 1 1 1)))                ;this signals that the phoneme is nil
         (e-list (cond ((= dif 0) (list s-1))
                       ((= dif 1) (list (append x s-1) (append s-1  x)))
                       ((= dif 2) (list (append x x s-1) (append x s-1 x) (append s-1 x x)))
                       ((= dif 3) (list (append x x x s-1) (append x x s-1 x) (append x s-1 x x) (append s-1 x x x)))
                       (t nil))))
  ;    (print e-list)
    (apply #'min (map 'list  #'(lambda (yy) (segment-distance-alt yy s-2)) e-list))))



;Stress can have the following values:
;0--a null syllable
;1--unstressed
;1.5--secondary stress
;2--primary stress
;stress-distance is the absolute difference divided by 2.
(defun stress-distance (s-1 s-2)
  (/ (abs (- s-1 s-2)) 2.0))

(defun cross-consonant-distance (onset coda)
  ;(print "Cross-consonant-distance")
  (let (
        (clean-onset (remove '(1 1 1 1 1 1 1 1 1 1 1 1 1) (remove '(0 0 0 0 0 0 0 0 0 0 0 0 0) onset)))
        )
    (segment-c-distance clean-onset coda)
        )
  )

;;;Calculates distance between two phonemes by the number of mismatched features
;;;Phoneme order does not matter.  Phonemes must be vectors of the same length

(defun phoneme-distance (phoneme-vector-1 phoneme-vector-2 &optional (null-distance
*average-consonant-distance*))
          (cond ((and (equal phoneme-vector-1 '(1 1 1 1 1 1 1 1 1 1 1 1 1))                             ;handles null cases
                      (equal phoneme-vector-2 '(1 1 1 1 1 1 1 1 1 1 1 1 1))) 0)
                ((or (equal phoneme-vector-1 '(1 1 1 1 1 1 1 1 1 1 1 1 1))
                     (equal phoneme-vector-2 '(1 1 1 1 1 1 1 1 1 1 1 1 1)))   null-distance)
                (t
                             (phoneme-distance-alt
                              phoneme-vector-1
                              phoneme-vector-2
                              0
                              (relevant-number-of-features phoneme-vector-1 phoneme-vector-2)))))






(defun phoneme-distance-alt (phoneme-vector-1 phoneme-vector-2 distance rel-num)
  (cond ((endp phoneme-vector-1) (/ distance
                                    rel-num))
        (t (phoneme-distance-alt (rest phoneme-vector-1)
                                 (rest phoneme-vector-2)
                                 (+ distance
                                    (if (equal (first phoneme-vector-1)
                                               (first phoneme-vector-2))
                                      0
                                      1))
                                 rel-num))))



;If word1 is shorter than word2, it adds the last phoneme of word2 to word1 repeatedly
; until they are of equal length,
;then returns the new word1.  If they are equal or word1 is longer, word1 is returned
;unchanged

(defun make-length-equal (word1 word2)
  (cond ((= (length word1) (length word2)) word1)
        ((> (length word1) (length word2)) word1)
        (t (make-length-equal (append word1 (last word1)) word2))))


(defun find-total-similarity (value-list &optional (total 0))
  (cond ((endp value-list) total)
        (t (find-total-similarity (rest value-list)
                                  (+ total
                                     (second (first value-list)))))))


(defun make-probability-list-sim (value-list)
  (make-probability-list-sim-alt value-list (find-total-similarity value-list) nil))

(defun make-probability-list-sim-alt (value-list total-sim probability-list)
  (cond ((endp value-list) probability-list)
        (t (make-probability-list-sim-alt (rest value-list)
                                          total-sim
                                          (append probability-list
                                                  (list (list (first (first value-list))
                                                              (/ (second (first value-list))
                                                                 total-sim))))))))



(defun relevant-number-of-features (phoneme-1 phoneme-2 &optional (irrelevant-number 0))
  (cond ((endp phoneme-1) (- 13 irrelevant-number))
        (t (relevant-number-of-features (rest phoneme-1)
                                        (rest phoneme-2)
                                        (+ irrelevant-number
                                           (if (and (eql (first phoneme-1) 'X)
                                                    (eql (first phoneme-2) 'X))
                                             1
                                             0))))))


;;;============================================================================
;;;   These Functions Conduct Experiment
;;;============================================================================


; (pprint (make-probability-list (make-value-list
;         (destroy-word
;          '((2 (X) (ae) (k))    (1 (s) (@) (X))   (1 (d) (@) (n t))) .1)
;         ph2)))


;Make an experiment that:    1. takes a list and a decay percent
;                            2.  For each word, runs N trials where probability-list is generated
;                            3.  Outputs the corresponding matrix of average probabilities





(defun n-trials (word list hazard-rate n)
  (average-trials (n-trials-alt word list hazard-rate n)))

(defun n-trials-alt (word list hazard-rate n &optional (results '()))
  (if (= n 0)
    results
    (n-trials-alt word list hazard-rate (- n 1) (append results
                                                        (list (trial word list hazard-rate))))))

(defun average-trials (trial-block)
  (let ((data-pack (make-data-matrix trial-block)))
    (list (first data-pack)
          (average-trials-alt (second data-pack)))))

(defun average-trials-alt (data-matrix)
  (let ((n (length data-matrix)))
    (map 'list
         #'(lambda (l) (roundoff (/ l n)))
         (map 'list
              #'(lambda (l) (apply #'+ l))
              (transpose data-matrix)
              ))
    ))


(defun strip-probability (list &optional (word-list nil) (prob-list nil))
  (if (endp list)
    (list word-list prob-list)
    (strip-probability (rest list)
                       (append word-list (list (first (first list))))
                       (append prob-list (list (second (first list)))))))


;This makes a two-part list, the first containing the list of words and the second containing
;;an n by m data matrix (nested list) where n is the number of trials and m is the number of words
;;  It takes as input the output from trials-alt

(defun make-data-matrix (trial-block)
  (list (first (strip-probability (first trial-block)))
        (make-data-matrix-alt trial-block nil)))

(defun make-data-matrix-alt (trial-block data-matrix)
  (if (endp trial-block)
    data-matrix
    (make-data-matrix-alt (rest trial-block)
                          (append data-matrix
                                  (list (second (strip-probability (first trial-block))))))))




(defun cross-list-trial (word-list hazard-rate trials)
  (list
   word-list
   (cross-list-trial-alt word-list word-list hazard-rate trials nil)))

(Defun Cross-list-trial-alt (Word-List total-list Hazard-Rate trials Data)
  (If (Endp Word-List) Data
      (Cross-list-trial-alt (Rest Word-List)
                            Total-List
                            Hazard-Rate
                            Trials
                            (Append Data
                                    (List (second (n-trials (First Word-List)
                                                            Total-List
                                                            Hazard-Rate
                                                            Trials
                                                            )))))))




(defun run-expt (word-list hazard-rate n)
  (let*
    ((matrix (second (cross-list-trial word-list hazard-rate n)))
     (diagonal (diag matrix))
     (l (length diagonal))
     (mean (expt (apply #'* diagonal)
                 (/ 1 l))))
    (list word-list
          matrix
          diagonal
          mean)))

;;;============================================================================
;;;   These Functions Evaluate Lists in Various Ways
;;;============================================================================

;takes a list and creates a matrix of pairwise similarities

(defun cross-evaluate-list (word-list &optional  (full-list word-list)  similarities)
  (cond ((endp word-list) similarities)
        (t (cross-evaluate-list (rest word-list)
                                full-list
                                (append similarities
                                        (list (cross-word-to-list (first word-list)
                                                                  full-list)))))))

;creates a vector of the similarities of a single word to an entire list
(defun cross-word-to-list (word word-list &optional similarities)
  (cond ((endp word-list) similarities)
        (t (cross-word-to-list word
                               (rest word-list)
                               (append similarities
                                       (list    (evaluate-similarity  (vectorify-word word)
                                                                      (first word-list))))))))

;;this is like cross-evaluate-list, but it keeps the 
;;similarities in a hierarchical word-based list structure

(defun cross-evaluate-list-list (word-list)
  ;(print "CROSS-EVALUATE-LIST-LIST")
     (mapcar #'(lambda (x) (cross-word-to-list-list x word-list)) word-list)) 




;creates a vector of the similarities of a single word to an entire list
(defun cross-word-to-list-list (word word-list)
  (mapcar #'(lambda (x) (evaluate-similarity-list (vectorify-word word) 
                                                  (vectorify-word x)))
    word-list))



(defun sum-matrix-list (matrix-list)
    (sum-vector-list
     (mapcar 'sum-vector-list matrix-list)))

(defun sum-vector-list (vector-list)
  (mapcar (lambda (x) (mapcar (lambda (y) 
                                (roundoff (apply '+ y) ))
                        (transpose x))) 
    (transpose vector-list))
  )


;this allows you to use the sum-vector-list and divide by list length
(defun divide-similarity-profile-by-n (profile n)
  (mapcar #'(Lambda (syllable) 
            (mapcar #'(lambda (y) (roundoff (/ y n))) syllable))
    profile))


(defun collapse-inner-entropy-lists (entropy-list)
  (mapcar #'(lambda (syllable) (mapcar #'(lambda (x) (apply '+ x)) syllable)) 
    entropy-list))
    
    
    
    
    (defun convert-sims-matrix (sim-matrix &optional (parameter 1) (prob-matrix nil))
  (cond ((endp sim-matrix) prob-matrix)
        (t (convert-sims-matrix (rest sim-matrix)
                                parameter
                                (append prob-matrix
                                        (list (convert-sims-vector (first sim-matrix) )))))))





;=============================================================================================
;          Finds distance between phonemes
;=============================================================================================
(defun pd (p1 p2)
  (phoneme-distance (vectorify-phoneme p1)
                    (vectorify-phoneme p2)))

(defun p-to-list (p plist &optional distances)
  (if (endp plist)
    distances
    (p-to-list p (rest plist) (append distances
                                      (list (pd p (first plist)))))))


(defun list-to-list (l1 l2 &optional matrix)
  (if (endp l1)
    matrix
    (list-to-list (rest l1)
                  l2
                  (append matrix (list (p-to-list (first l1)
                                                  l2))))))

(defun list-to-list-average (l1 l2)
  (roundoff (/ (sum-list (list-to-list l1 l2))
     (* (length l1)
        (length l2))
        )))
  

;==================================================================
;     Processes words and puts them into the correct template
;==================================================================

;;this process-onset has no effect on the onset and is shadowed by the next defun which does.
;;it should only be used for testing purposes

;(defun process-onset (onset)
;               onset)
 
(defun process-onset (onset)
  (let* ((onset-a '(s z sh zh))
         (onset-b '(p k t h b d ch j g f v th thz m n))
         (onset-c '(y w r l))
         (testa    (member (first onset)   onset-a))
         (rest1    (if testa (rest onset) onset))                  ;this is the remaining critical string
         (new-a    (if testa (first onset) 'X))
         (testb    (member (first rest1)  onset-b))
         (rest2    (if testb (rest rest1) rest1))
         (new-b    (if testb (first rest1) 'X))
         (new-c    (if (member (first rest2) onset-c) (first rest2) 'X)))
    (list new-a new-b new-c)))

(defun process-vowel (vowel)
  (if (= (length vowel) 1)
      (append vowel vowel)
      vowel))
 
(defun process-syllable (syllable)
  (list (first syllable)
        (process-onset (second syllable))
        (process-vowel (third syllable))
        (fourth syllable)))

(defun process-word (word &optional processed-word)
  (if (endp word)
    processed-word
    (process-word (rest word) (append  processed-word
                                       (list (process-syllable (first word)))))))
(defun p-l (list &optional new-list)
  (if (endp list)
      new-list
      (p-l (rest list)
           (append new-list
                   (list (process-word (first list)))))))

;==================================================================
;     Processes words and puts them into the correct template
;      for the purpose of measuring entropy
;==================================================================
;
;These mimic earlier functions, but align templates in the nucleus and coda
; in order to better measure entropy


(defun process-nucleus (phoneme-cluster)
  (if (= (length phoneme-cluster) 2)
      phoneme-cluster
    (append phoneme-cluster (list (first phoneme-cluster)) )));just double the current vowel


;I don't want to deal with this, so I will just add null phonemes on the end until the length is 4

 (defun process-coda (phoneme-cluster)
   (if (= (length phoneme-cluster) 4)
       phoneme-cluster
     (process-coda (append phoneme-cluster '(X) ))))
   
   

(defun process-syllable-e (syllable)
  (list (first syllable)
        (process-onset (second syllable))
        (process-nucleus (third syllable))
        (process-coda    (fourth syllable))))

(defun process-word-e (word &optional processed-word)
  (if (endp word)
    processed-word
    (process-word-e (rest word) (append  processed-word
                                       (list (process-syllable-e (first word)))))))
(defun p-l-e (list &optional new-list)
  (if (endp list)
      new-list
      (p-l-e (rest list)
           (append new-list
                   (list (process-word-e (first list)))))))

;======================================================================
;  Information-theory based Weighting of syllable contents
;======================================================================

;;In an attempt to determine how to weight the different sections of the 
;;phonemes in a set of words, we find the information being carried by that set of
;;words.

;;Entropy is the calculated for each onset, rhyme, and coda in a set of words.
;;This requires "Transposing" the word-set hierarchy.  The tricky part is to 

;;parts with different numbers of phonemes.

;;onsets have a template, so they are not too difficult.



;;Each feature has one of four values---  +, -, X (no value in phoneme), and 0 (no phoneme)




;;This takes a list of symbols +-x0 and generates a list of 4 probabilities, designating how 
;;frequently each appears in the list.

;;;for this function to work properly, the LIST-OF-FEATURES must include only the four symbols below
;;;It will then spit out a list of probabilities of each symbol.





(defun calculate-probabilities (list-of-features)  ;divides the feature-counts by the total number
  (mapcar #'(lambda (x) 
              (if (= x 0)
                  0.0             ;either it is 0 or a probability--this avoids dividing by 0
                (/ x (- (length list-of-features) 0)) 
               ;       (count 'x list-of-features)))
              ))
    (list (count '+ list-of-features)   ;This counts the number of each feature
          (count '- list-of-features)
          (count 'X list-of-features)
          (count '1 list-of-features)
          )))



;;This calculates entropy (-Sum p log_2_p for a column of features.
;; Since log of 0 is undefined, special care must be taken to make sure
;; 0 log 0 is counted as 0.

(defun calculate-entropy (list-of-probabilities &optional (sum 0))
  (if (endp list-of-probabilities) (- sum)
    (calculate-entropy (rest list-of-probabilities)
                       (let ((p (first list-of-probabilities)))
                         (+ sum
                          (if (= p 0) 
                                 0.0
                            (* p (log p 2))))))))






(defun pl-1 (l)
  (list (do-list-of-syllables (transpose (first (transpose (vectorify-list (p-l-e l))))))))

;;;this only works for 1-syllable words
;;; use:  (pl-1 list-of-words)


(defun pl-2 (l)
  (let*  ((l (transpose (vectorify-list (p-l-e l))))
          (l1 (transpose (first l)))
          (l2 (transpose (second l))))
    (List
     (do-list-of-syllables l1)
     (do-list-of-syllables l2)
     )
    )
  )



(defun pl-3 (l)
  (let*  ((l (transpose (vectorify-list (p-l-e l))))
          (l1 (transpose (first l)))
          (l2 (transpose (second l)))
          (l3 (transpose (third l)))
          )
    (List
     (do-list-of-syllables l1)
     (do-list-of-syllables l2)
     (do-list-of-syllables l3)
     )
    )
  )



(defun do-list-of-syllables (l)
  (let* 
      ((tran-l   l)  ;transposed list
       (weights (first tran-l))
       (onsets  (second tran-l))
       (nuclei  (third tran-l))
       (codas   (fourth tran-l)))
    (list
    (mapcar #'(lambda (x) (apply #'+       ;calculates entropy for onset phonemes
                                 (mapcar #'calculate-entropy 
                                        (Mapcar #'calculate-probabilities (transpose x)))))
      (transpose onsets) )
     
     (mapcar #'(lambda (x) (apply #'+      ;calculates entropy for nuclei
                                 (mapcar #'calculate-entropy 
                                        (Mapcar #'calculate-probabilities (transpose x)))))
      (transpose nuclei) )
     
  
   (mapcar #'(lambda (x) (apply #'+      ;calculates entropy for codas
                                  (mapcar #'calculate-entropy 
                                        (Mapcar #'calculate-probabilities (transpose x)))))
      (transpose codas) )

     
     )
     
     ))
  

