;; PSIMETRICA:  Phonological SIMilarity METRIC Analysis
;; (c) 1998-2004 Shane T. Mueller
;; stmuelle@indiana.edu
;; More info at: http://mypages.iu.edu/~stmuelle
;; May not be copied or distributed without permission.
;; Tools and routines for measuring phonological similarity between words.
;;
;;  tversky.lsp
;;  This file contains utilities for measuring phonological similarity, based
;;  on a Tversky features-of-similarity concept.  This may or may not work
;;  with the current setup.


;;THIS SHADOWS ABOVE DEFINITION AND IS SHADOWED BELOW--IT MEASURES SIMILARITY in a tversky-esque manner.
;; tHIS may avoid some of the arbitrariness involved when you match null phonemes against actual phonemes

;;traditionally, the features of tversky are unary not binary, so it may take a little rearranging.
;;  There are four potentially interesting quantities:
;;                               common features features--both words take the same value on a features
;;                               opposing features       --words take opposite values on a feature
;;                               A-B                     --A takes a value on this feature while B does not
;;                               B-A                     --B takes a value on this feature while A does not

;; This should be done for each syllable independently, averaging can then be done later
;; if a phoneme has X for a feature, it should be the same as if there is no phoneme
;;good luck

;;This assumes that the words have equal numbers of syllables.
;; and that all phoneme clusters are of equal length as well.
;; this requires a little arrangement when the codas differ in length
;;To handle this, phoneme-cluster-similarity-tversky handles this

;;This would be a really good place to use classes


(defvar *coinciding-weight* 1)
(defvar *opposing-weight* 1)
(defvar *absence-weight* 1)

(setq *coinciding-weight* 1)
(setq *opposing-weight* 1)
(setq *absence-weight* 1)


(defun evaluate-similarity-alt (word-feature-vector target-word-vector similarity num-syllables)
  (mapcar #'(lambda (x y) (syllable-similarity-tversky x y)) 
    word-feature-vector 
    target-word-vector))


(defun syllable-similarity-tversky (syllable-1 syllable-2)
  (mapcar #'phoneme-cluster-similarity-tversky
    (rest syllable-1)
    (rest syllable-2)))


;assumes that onset is already aligned with p-l algorithms, so that it fits into standard template.
;Also, assumes that the nucleus is doubled up if necessary.
;;trouble comes in the coda
;;these should by dynamically rearranged so that similarity is maximazed.
;;For now, just assume that blanks have been added to the end so that all codas are 4 long.

(defun phoneme-cluster-similarity-tversky (phoneme-cluster-1 phoneme-cluster-2)
  (apply '+ (mapcar #'phoneme-similarity-tversky  phoneme-cluster-1 phoneme-cluster-2)))

;  this takes two phonemes and an (optional) similarity list that contains:
;;  (coinciding opposing absence12 absence21)

(defun phoneme-similarity-tversky (phoneme-1 phoneme-2)
  (phoneme-similarity-tversky-alt phoneme-1 phoneme-2 '(0 0 0 0)))

(defun phoneme-similarity-tversky-alt (phoneme-1 phoneme-2 similarity)
  (if (endp phoneme-1)
      (dot (list (* 1 *coinciding-weight*)
                 (* -1 *opposing-weight*)
                 (* -1 *absence-weight* )
                 (* -1 *absence-weight*))
           similarity)
    (let ((f1 (first phoneme-1))
          (f2 (first phoneme-2))
          (c (first similarity))
          (o (second similarity))
          (a1-2 (third similarity))
          (a2-1 (fourth similarity))
          )
      (if (and (eql f1 (or '+ '-)) (eql f2 (or '+ '-)))
          (if (eql f1 f2) (incf c) (incf o))
        (if (eql f1 (or '+ '-)) (incf a1-2)
          (if (eql f2 (or '+ '-)) (incf a2-1))))
      (phoneme-similarity-tversky-alt (rest phoneme-1) (rest phoneme-2) (list c o a1-2 a2-1)))))



;This just subtracts the diagonal entries from each row, giving a distance matrix
(defun make-distance-matrix-from-tversky (matrix)
  (mapcar #'(lambda (row diagonal-entry)  
              (mapcar #'(lambda (x)  (subtract-words diagonal-entry x)) row))
    matrix
    (diag matrix)))


;;take a pair of word distances: ( ( # # # ) ( # # # ) ( # # # ) )
;; and finds the difference

(defun subtract-words (word1 word2)
  (mapcar #'(lambda (syll1 syll2)
              (mapcar #'(lambda (ph1 ph2) (if (= ph1 0) 0 (/ (- ph1 ph2) ph1)) 
                syll1 syll2))
              word1 word2))
  )